









if (!global.$apollo) {



/**
 * Extend an object with another object
 * @param  {Object} obj      object to be extended
 * @param  {Object} ext      extension object
 * @param  {bool} override   Overwrite existing properties in obj
 * @param  {bool} deep       Doing an deep extend (perform extend on every object property)
 * @return {Object}          reference to obj
 */
function $extend(obj, ext, override, deep) {
  var key;
  if (override) {
    if (deep)
      _overrideDeepExtend(obj, ext);
    else
      for (key in ext)
        obj[key] = ext[key];
  } else {
    if (deep)
      _deepExtend(obj, ext);
    else
      for (key in ext)
        if (!(key in obj))
          obj[key] = ext[key];
  }
  return obj;
}

function _overrideDeepExtend(obj, ext) {
  for (var key in ext)
    if (Object.isObjectStrict(obj[key]) && Object.isObjectStrict(ext[key]))
      _overrideDeepExtend(obj[key], ext[key]);
    else
      obj[key] = ext[key];
}

function _deepExtend(obj, ext) {
  for (var key in ext)
    if (Object.isObjectStrict(obj[key]) && Object.isObjectStrict(ext[key]))
      _deepExtend(obj[key], ext[key]);
    else if (!(key in obj))
      obj[key] = ext[key];
}

/**
 * Define properties of an Object, Which usually used to extend prototype
 *   of an object, as it will set properties as non-enumerable, and will
 *   turn setValue(value) and getValue() functions to setter and getters.
 * Note: You should only use $define or Object.defineProperty on prototype,
 *   or on a class' itself (to define static methods), instead of on instances
 *   which could lead to severe performance issue.
 * @param  {Object} object    target object
 * @param  {Object} prototype extension object
 * @param  {bool} preserve    preserve existing property
 * @return {Object}           reference to object
 */
function $define(object, prototype, preserve) {
  Object.getOwnPropertyNames(prototype).forEach(function(key) {
    if (preserve && (key in object))
      return;
    var desc = Object.getOwnPropertyDescriptor(prototype, key);
    if ('value' in desc)
      desc.writable = true;
    delete desc.enumerable;
    delete desc.configurable;
    Object.defineProperty(object, key, desc);
  });
  return object;
}

/**
 * Declare a Class.
 * @param  {Function} fn      constructor of the Class
 * @param  {Object} prototype prototype of Class
 * @return {Function}         reference to constructor
 */
function $declare(fn, prototype) {
  fn.prototype.constructor = fn;
  $define(fn.prototype, prototype);
  return fn;
}

/**
 * Inherit another Class to current Class
 * @param  {Function} fn      constructor of the Class
 * @param  {Function} parent  parent Class
 * @param  {Object} prototype prototype of Class
 * @return {Function}         reference to constructor
 */


function $inherit(fn, parent, prototype) {
  fn.prototype = {
    constructor: fn,
    __proto__: parent.prototype
  };
  if (prototype)
    $define(fn.prototype, prototype);
  return fn;
}



/**
 * Adding enumerations to a Class (both static and prototype).
 * @param  {Function} fn     constructor of the Class
 * @param  {Object}   values object holding all enumerates want to define
 * @return {Function}        reference to constructor
 */
function $defenum(fn, values) {
  $define(fn, values);
  $define(fn.prototype, values);
  return fn;
}

/**
 * Format a string with given pattern.
 * @param  {string} str pattern
 * @return {string}     formatted string
 */


var $format = require('util').format;



/**
 * Making an Error instance with given format and parameters.
 * Note: this is a helper function works like util.format(),
 *   apart from it returns an Error object instead of string.
 * @return {Error} generated Error instance
 */
function $error() {
  return new Error($format.apply(null, arguments));
}


/**
 * Generate a deep copy of an Object with its primitive typed
 * fields (exclude functions).
 * @param  {mixed} obj  source object
 * @return {mixed}      cloned object
 */
function $valueCopy(obj) {
  var res;
  if (Array.isArray(obj)) {
    res = obj.slice(0);
    for (var i = 0; i < res.length; i++)
      if (Object.isObject(res[i]))
        res[i] = $valueCopy(res[i]);
  } else if (Object.isObjectStrict(obj)) {
    res = {};
    for (var key in obj)
      res[key] = $valueCopy(obj[key]);
  } else if (Function.isFunction(obj)) {
    return undefined;
  } else {
    return obj;
  }
  return res;
}

/**
 * Generates a copy of an Object.
 * @param  {Mixed} org  source object
 * @param  {bool} deep  perform a deep clone
 * @return {Mixed}      cloned object
 */
function $clone(obj, deep) {
  var res;
  var _deep = deep === true || deep - 1;
  if (Array.isArray(obj)) {
    res = obj.slice(0);
    if (deep)
      for (var i = 0; i < res.length; i++)
        if (Object.isObject(res[i]))
          res[i] = $clone(res[i], _deep);
  } else if (Object.isObjectStrict(obj)) {
    res = {};
    for (var key in obj)
      res[key] = obj[key];
    if (deep)
      for (var key in obj)
        if (Object.isObject(res[key]))
          res[key] = $clone(res[key], _deep);
  } else {
    return obj;
  }
  return res;
}

/**
 * Return default value of an undefined variable.
 * @param  {Mixed} val  value
 * @param  {Mixed} def  default value
 * @return {Mixed}
 */
function $default(val, def) {
  return val === undefined ? def : val;
}

/**
 * Wrap an object with given Class.
 * Note: it will call Class.__wrap method to do custom wrapping.
 * @param  {Object} obj     object to be wrapped
 * @param  {Function} Type  wrapping Class
 * @return {Object}         wrapped object
 */


function $wrap(obj, Type) {
  obj.__proto__ = Type.prototype;
  if (Type.__wrap)
    Type.__wrap(obj);
  return obj;
}

/**
 * Removing prototype chain from a given object.
 * @param  {Object} object   object to be stripped
 * @return {Object}          object stripped
 */
function $strip(object) {
  object.__proto__ = Object.prototype;
  return object;
}

/**
 * Use Object.prototype.toString to determine an element's type
 * This method provide more stricter strategy on type detection,
 * can be worked with typeof.
 * @param  {Mixed}  obj  Variable
 * @return {String}      type of the variable, like typeof,
 *                       but with better precision.
 */
function $typeof(obj) {
  var type = Object.prototype.toString.call(obj);
  return type.substring(8, type.length - 1).toLowerCase();
}



$define(global, {
  $extend: $extend,
  $define: $define,
  $declare: $declare,
  $inherit: $inherit,
  $defenum: $defenum,
  $format: $format,
  $error: $error,
  $valueCopy: $valueCopy,
  $clone: $clone,
  $default: $default,
  $wrap: $wrap

  ,
  $apollo: require('./package').version,
  $strip: $strip,
  $typeof: $typeof

});

$define(String.prototype, {
  /**
   * Repeat current string for given times.
   * @param  {int} times  Times to repeat
   * @return {string}     result
   */
  repeat: function(times) {
    var res = '';
    for (var i = 0; i < times; i++)
      res += this;
    return res;
  },
  /**
   * Padding this to given length with specified char from left.
   * @param  {char} ch    padding char
   * @param  {int} length desired length
   * @return {string}     result
   */
  paddingLeft: function(ch, length) {
    if (this.length < length)
      return ch.repeat(length - this.length) + this;
    return this;
  },
  /**
   * Padding this to given length with specified char from right.
   * @param  {char} ch    padding char
   * @param  {int} length desired length
   * @return {string}     result
   */
  paddingRight: function(ch, length) {
    if (this.length < length)
      return this + ch.repeat(length - this.length);
    return this;
  },

  /**
   * get last character in this string
   * @return {String} last character
   */
  get back() {
    return this[this.length - 1];
  },
  /**
   * get first character in this string
   * @return {String} first character
   */
  get front() {
    return this[0];
  },

  /**
   * Tests if this string starts with the given one.
   * @param  {string} str string to test with
   * @param  {number} pos optional, position to start compare, defaults
   *                      to 0
   * @return {bool}       result
   */
  startsWith: function(str, pos) {
    if (str === null || str === undefined || str.length === 0)
      return true;
    return this.substr(pos || 0, str.length) === str;
  },
  /**
   * Tests if this string ends with the given one.
   * @param  {string} str string to test with
   * @param  {number} len optional, pretend this string is of given length,
   *                      defaults to actual length
   * @return {bool}       result
   */
  endsWith: function(str, len) {
    if (str === null || str === undefined || str.length === 0)
      return true;
    return this.substr((len || this.length) - str.length, str.length) === str;
  },
  /**
   * Return a string in it's title form.
   * @return {string} string in title case
   * Note: if a word containing upper case, nothing
   *   will be done.
   */
  toTitleCase: function() {
    return this.replace(/\b([a-z])(['a-z]*)\b/g, function(all, letter, rest) {
      return letter.toUpperCase() + rest;
    });
  },
  /**
   * Trim whitespaces at the begining of the string
   * @return {string} trimmed string
   */
  trimLeft: function() {
    return this.replace(/^\s+/, '');
  },
  /**
   * Trim whitespaces at the ending of the string
   * @return {string} trimmed string
   */
  trimRight: function() {
    return this.replace(/\s+$/, '');
  }
}, true);

$define(Number.prototype, {
  /**
   * Clamp current value to the given range [lb, ub]
   * @param  {number} lb lower bound
   * @param  {number} ub upper bound
   * @return {number}    result
   */
  clamp: function(lb, ub) {
    var rtn = Number(this);
    if (lb !== undefined && rtn < lb)
      rtn = lb;
    if (ub !== undefined && rtn > ub)
      rtn = ub;
    return rtn;
  },
  /**
   * Shortcut to Math.floor(this)
   * @return {number} Math.floor(this)
   */
  floor: function() {
    return Math.floor(this);
  },
  /**
   * Shortcut to Math.ceil(this)
   * @return {number} Math.ceil(this)
   */
  ceil: function() {
    return Math.ceil(this);
  },
  /**
   * Shortcut to Math.round(this) with additional parameters
   * @param  {number} decimals number of decimal digits to round up to
   * @return {number}          rounded number
   */
  round: function(decimals) {
    if (decimals) {
      var unit = Math.pow(10, decimals);
      return Math.round(this * unit) / unit;
    }
    return Math.round(this);
  },
  /**
   * Get the thousands separated number
   * @param  {number} decimals  number of decimal digits to remain
   * @param  {string} separator separator
   * @return {string}           separated number
   */
  toGroup: function(decimals, separator) {

    decimals = decimals || 0;

    if (this > -1000 && this < 1000)
      return this.toFixed(decimals);

    separator = separator || ',';

    var sign = this < 0 ? '-' : '';
    var tmp = Math.abs(this).toFixed(decimals);

    var intPart, decimalPart;
    if (decimals > 0) {
      intPart = tmp.substr(0, tmp.length - decimals - 1);
      decimalPart = tmp.substr(tmp.length - decimals - 1);
    } else {
      intPart = tmp;
      decimalPart = '';
    }

    var res = '';
    for (var pos = 0, len = intPart.length % 3 || 3;
        pos < intPart.length; pos += len, len = 3) {
      if (res !== '')
        res += separator;
      res += intPart.substr(pos, len);
    }
    return sign + res + decimalPart;

  }
});

$define(Array.prototype, {
  /**
   * get minimum value in this array
   * @return {Mixed} minimal value
   */
  min: function() {
    var res = this[0];
    for (var i = 1; i < this.length; i++)
      if (this[i] < res)
        res = this[i];
    return res;
  },
  /**
   * get maximum value in this array
   * @return {Mixed} maximum value
   */
  max: function() {
    var res = this[0];
    for (var i = 1; i < this.length; i++)
      if (this[i] > res)
        res = this[i];
    return res;
  },
  /**
   * Push a value iif it's not in this array, and return value's index.
   * @param  {Mixed} val  new value
   * @return {int}        index of the value
   * Note: This only works with primitive typed elements, which can be found
   *       with Array#indexOf().
   */
  add: function(val) {
    var index = this.indexOf(val);
    if (index === -1)
      return this.push(val) - 1;
    return index;
  },
  /**
   * Find a value in the array and remove it.
   * @param  {Mixed} val  value to remove
   * @return {Array}      this
   * Note: This only works with primitive typed elements, which can be found
   *       with Array#indexOf().
   */
  remove: function(val) {
    var index = this.indexOf(val);
    if (index > -1) {
      // Shift copy elements instead of Array#splice() for better performance.
      // http://jsperf.com/fast-array-splice/18
      while (++index < this.length)
        this[index - 1] = this[index];
      this.pop();
    }
    return this;
  },
  /**
   * Rotate this array (n->0, n+1->1, ...)
   * @param  {int} n   the offset
   * @return {Array}   this
   */
  rotate: function(n) {
    if (n < 0)
      n = n % this.length + this.length;
    n %= this.length;
    var middle = n;
    var next = n;
    var first = 0;
    while (first < this.length) {
      var t = this[first];
      this[first] = this[next];
      this[next] = t;
      first++;
      next++;
      if (next == this.length) next = middle;
      else if (first == middle) middle = next;
    }
    return this;
  },

  /**
   * get last element in this array
   * Note: It's not a reference when returning a non-object!
   * @return {Mixed} last element
   */
  get back() {
    return this[this.length - 1];
  },
  /**
   * get first element in this array
   * Note: It's not a reference when returning a non-object!
   * @return {Mixed} first element
   */
  get front() {
    return this[0];
  },

  /**
   * Flattern a array with sub arrays.
   * @param  {bool} deep if continue to flatten sub arrays
   * @return {Array}     flattened array.
   */
  flatten: function(deep) {
    var res = [];
    if (!deep)
      return res.concat.apply(res, this);
    for (var i = 0; i < this.length; i++)
      if (Array.isArray(this[i]))
        res.push.apply(res, this[i].flatten(true));
      else
        res.push(this[i]);
    return res;
  },
  /**
   * Return unique elements in the array
   * @return {Array}
   */
  unique: function() {
    var res = [];
    var dict = {};
    for (var i = 0; i < this.length; ++i) {
      var key = this[i].toString();
      if (dict.hasOwnProperty(key))
        continue;
      dict[key] = true;
      res.push(this[i]);
    }
    return res;
  },
  /**
   * shuffle elements in the array in-place
   * @return {Array}
   */
  shuffle: function() {
    for (var n = this.length; n > 0; n--) {
      var idx = Math.floor(n * Math.random());
      if (idx != n - 1) {
        var tmp = this[idx];
        this[idx] = this[n - 1];
        this[n - 1] = tmp;
      }
    }
    return this;
  }
});

/**
 * Forward declaring prototype functions to Array's static
 * methods.
 */
if (Array.map === undefined)
  ['forEach', 'every', 'some', 'filter', 'map', 'reduce', 'reduceRight', 'slice']
    .forEach(function(method) {
      var fn = Array.prototype[method];
      Object.defineProperty(Array, method, {
        value: function(a, b, c) {
          return fn.call(a, b, c);
        }
      });
    });

if (String.trim === undefined)
  ['trim', 'trimLeft', 'trimRight']
    .forEach(function(method) {
      var fn = String.prototype[method];
      Object.defineProperty(String, method, {
        value: function(a) {
          return fn.call(a);
        }
      });
    });

$define(Object, {
  /**
   * Determine if an object is empty
   * @param  {Object} obj  object to test
   * @return {bool}        object is empty
   */
  isEmpty: function(obj) {
    if (!obj)
      return true;
    for (var key in obj)
      return false;
    return true;
  },
  /**
   * Get values of an object, like Object.keys().
   * @param  {Object} obj  object to extract
   * @return {Array}       values in the object
   */
  values: function(obj) {
    return Object.keys(obj).map(function(k) {
      return obj[k];
    });
  },
  /**
   * Vague but fast isObject test
   * Note: new String(), function, array, etc will return true
   * @param  {Mixed} obj  object to test
   * @return {bool}       true if obj is an object and not null
   */
  isObject: function(obj) {
    /**
     * Known fastest way to test, the order of the test
     * following: http://jsperf.com/typeof-vs-bool.
     */
    return obj && typeof obj === 'object';
  },
  /**
   * Strict isObject test, only pure Object will return true
   * Note: only {} will return true
   * @param  {Mixed} obj  object to test
   * @return {bool}       true if obj is strictly an object
   */
  isObjectStrict: function(obj) {
    return Object.prototype.toString.call(obj) === '[object Object]';
  }

  ,
  /**
   * project $object with projectiong, same behaviour with mongodb projection
   * @param  {Object} object      target object
   * @param  {Object} projection  An object mapping fields to values
   * @param  {Boolean} deep       if true, go deep for sub objects
   * @param  {Boolean} keep       if true, keep undefined field of this
   * @return {Object}             projected object
   */
  project: function(object, projection, deep, keep) {
    if (!Object.isObject(projection))
      return object;
    var res = {};
    Object.keys(projection).forEach(function(key) {
      var proj = projection[key];
      if (proj) {
        var obj = object[key];
        if (deep && Object.isObjectStrict(obj) && Object.isObjectStrict(proj)) {
          res[key] = Object.project(obj, projection[key], deep, keep);
        } else {
          if (keep)
            res[key] = obj;
          else if (obj !== undefined)
            res[key] = obj;
        }
      }
    });
    return res;
  },
  Transformer: function(mapping) {
    var expr = [];
    expr.push('exec=function (object) {');
    expr.push('var res = {};');
    (function loop(lhv, mapping) {
      Object.keys(mapping).forEach(function(key) {
        var source = mapping[key];
        if (/\W/.test(key)) key = '["' + key + '"]';
        else key = '.' + key;


        var target = lhv + key;
        if ($typeof(source) == 'object') {
          expr.push(target + ' = {};');
          return loop(target, source);
        }

        if (true === source)
          source = 'object' + key;
        else if ($typeof(source) == 'string')
          source = 'object' + source;
        else if ($typeof(source) == 'function')
          source = '('+source.toString()+')(object)';
        expr.push(target + ' = ' + source + ';');
      });
    })('res', mapping);
    expr.push('return res;');
    expr.push('}');
    this.exec = eval(expr.join(''));
  }

});

$define(Function, {
  /**
   * Test if an object is a function
   * @param  {Mixed} obj  object to test
   * @return {bool}       true if so
   */
  isFunction: function(obj) {
    return typeof obj === 'function';
  }
});

$define(Date, {
  /**
   * Cast a value to Date
   * @param  {Mixed} obj  object to cast
   * @return {Date}       casted value
   */
  cast: function(obj) {
    if (obj instanceof Date)
      return obj;
    if (typeof obj === 'string')
      obj = Date.parse(obj);
    if (typeof obj === 'number') {
      if (isNaN(obj))
        return null;
      obj = new Date(obj);
      if (isNaN(obj.valueOf()))
        return null;
      return obj;
    }
    return null;
  },
  /**
   * Determine if an object is a Date
   * @param  {Object}     object to test
   * @return {bool}       true iif it's a date.
   */
  isDate: function(obj) {
    obj = Date.cast(obj);
    if (obj)
      return obj >= 0 && obj < 2147483647000;
    return false;
  }
});

$define(Boolean, {
  /**
   * Cast a value to bool
   * @param  {Object} obj  object to cast
   * @return {bool}        casted value
   */
  cast: function(obj) {
    if (obj === true || obj === false)
      return obj;
    if (typeof obj === 'string')
      return (/^(true|yes|ok|y|on)$/i).test(obj);
    return Boolean(obj);
  }
});

$define(RegExp, {
  /**
   * Escape a string to work within a regular expression
   * @param  {string} str string to escape
   * @return {strign}     escaped string
   */
  escape: function(str) {
    return str.replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&');
  }
});

$define(JSON, {
  /**
   * Try to parse a json string
   * @param  {string} str json string
   * @return {mixed}      parsed result
   */
  tryParse: function(str) {
    try {
      return JSON.parse(str);
    } catch(e) {
      return;
    }
  }
});





}



